from pathlib import Path
import os
import pytest


@pytest.fixture(autouse=True)
def prompt2_env(request, tmp_path):
    path = Path(request.fspath)
    fixtures_path = path.parent / 'fixtures/prompt2'
    env = dict(
        PROMPT2_CACHE_PATH=str(fixtures_path / 'cache'),
        PROMPT2_USER_PATH=str(tmp_path / 'prompts_user'),
        PROMPT2_LOCAL_PATH=str(tmp_path / 'prompts_local'),
        EDITOR='cat',
        NO_TIMESTAMPER='1',
    )
    os.environ.update(env)
    return env



"""
Juggle between multiple models by defining environment variables.

.. envvar:: MODEL

    An environment variable to configure the default LLM plugin, as far as
    the base plugin is concerned, it must be in the following format::

        [<code2 backend plugin>] [<args>...] [<kwarg=value>...]

    When the backend plugin is omited, or not found,
    :py:class:`~code2.plugins.backend.litellm` is used by default which should
    be fine in most cases, leveraging the env vars you already have say for
    aider or anything using litellm, such as ``$OPENAI_API_KEY``,
    ``$OPENROUTER_API_KEY``, etc. For the litellm backend, the first arg is the
    model name and kwargs are optional.

    The default value for MODEL env var is::

        MODEL=openrouter/google/gemini-2.5-pro-exp-03-25:free

    You can add litellm kwargs as such::

        MODEL='openrouter/deepseek/deepseek-chat:free max_tokens=16384'

.. envvar:: MODEL_name

    ``MODEL_name`` is not an actual variable, it represents other models
    configurations that you can have, ie::

        MODEL_EDITOR='deepseek/deepseek-chat:free max_tokens=16384'
        MODEL_ARCHITECT=xai/grok-2-latest

    While you will be able to use the model names you want in your own code2
    workflow plugins, code2 core prompt plugins will try to use the best one
    that is defined for a purpose based on standard names, "editor" and
    "architect" at this time.
    https://aider.chat/2024/09/26/architect.html
"""

from pathlib import Path
import cli2
import importlib.metadata
import json
import hashlib
import os

from cli2.exceptions import NotFoundError
from .plugins.litellm import LiteLLMPlugin
from .parser import Parser
from .plugin import Plugin


class Model:
    default = 'openrouter/deepseek/deepseek-r1:free'

    class NotFoundError(NotFoundError):
        title = 'MODEL NOT FOUND'

        def available_list(self):
            models = {
                key[6:].lower(): os.environ[key]
                for key in os.environ
                if key.startswith('MODEL_')
            }
            models['default'] = os.getenv('MODEL', Model.default)
            return models

    def __init__(self, backend=None):
        if not isinstance(backend, Plugin):
            configuration = self.configuration_get(backend or '')
            backend = self.backend_factory(configuration)
        self.backend = backend
        self._cache_path = None

    @property
    def cache_path(self):
        if not self._cache_path:
            if 'PROMPT2_CACHE_PATH' in os.environ:
                self._cache_path = Path(os.getenv('PROMPT2_CACHE_PATH'))
            else:
                self._cache_path = Path(os.getenv('HOME')) / '.code2/cache'
        return self._cache_path

    @cache_path.setter
    def cache_path(self, value):
        self._cache_path = value

    @classmethod
    def configuration_get(cls, name, strict=False):
        if not isinstance(name, (tuple, list, set)):
            name = [name]

        keys = [
            f'MODEL_{name.upper()}'
            for name in name
        ]

        key = None
        for _ in keys:
            if _ in os.environ:
                key = _
                break

        if not key:
            if strict:
                raise cls.NotFoundError(', '.join(name))
            if 'MODEL' in os.environ:
                key = 'MODEL'

        if key and key in os.environ:
            configuration = os.environ[key]
        else:
            # default value
            configuration = cls.default

        return configuration

    @staticmethod
    def configuration_parse(tokens):
        args = list()
        kwargs = dict()
        for token in tokens:
            key = None
            if '=' in token:
                key, value = token.split('=')
            else:
                value = token

            try:
                value = float(value)
            except ValueError:
                try:
                    value = int(value)
                except ValueError:
                    pass

            if key:
                kwargs[key] = value
            else:
                args.append(value)
        return args, kwargs

    @classmethod
    def backend_factory(cls, configuration):
        tokens = configuration.split()

        # load the backend plugin based on first token, litellm by default
        plugins = importlib.metadata.entry_points(
            name=tokens[0],
            group='prompt2',
        )
        if plugins:
            tokens = tokens[1:]
            plugin = [*plugins][0].load()
        else:
            plugin = LiteLLMPlugin

        args, kwargs = cls.configuration_parse(tokens)

        return plugin(*args, **kwargs)

    async def completion(self, messages, cache_key=None, stream=False):
        cache_key = cache_key or self.hash(
            messages + [str(self.backend)]
        )
        cache_path = self.cache_path / f'{cache_key}_response.txt'

        if cache_path.exists():
            cli2.log.debug('cache hit!', cache_path=cache_path, json=messages)
            with cache_path.open('r') as f:
                content = f.read()
            if stream:
                print(cli2.highlight(content, 'Markdown'))
            return content

        response = await self.backend.completion(messages)

        self.cache_path.mkdir(exist_ok=True, parents=True)
        cli2.log.debug('cached', cache_key=cache_key, response=response)
        with cache_path.open('w') as f:
            f.write(response)

        return response

    def hash(self, messages):
        _ = hashlib.new('sha1')
        cache = messages
        _.update(json.dumps(cache).encode('utf8'))
        return _.hexdigest()

    def parser(self, name):
        return Parser.get(name)(self)

    async def process(self, messages, parser=None, cache_key=None,
                      stream=False):
        if parser:
            messages = parser.messages(messages)

        tokens = sum([len(msg['content']) for msg in messages])
        cli2.log.debug('messages', tokens=tokens, json=messages)
        result = await self.completion(messages, stream=stream)
        cli2.log.debug('response', response=result)

        if parser:
            result = parser.parse(result)
            cli2.log.debug('parsed', result=result)

            class ParserResult(type(result)):
                pass

            _result = ParserResult(result)
            # attach parser and actual result
            _result.parser = parser
            _result.result = result
            result = _result
        return result

    async def __call__(self, messages, parser=None, cache_key=None,
                       stream=True):
        if isinstance(parser, str):
            # load parser object from string
            parser = self.parser(parser)

        if callable(getattr(messages, 'messages', None)):
            # quacks like a Prompt object
            messages = await messages.messages()

        return await self.process(messages, parser, cache_key=cache_key,
                                  stream=stream)

    def __str__(self):
        return str(self.backend)



import cli2
import flow2
from prompt2 import Model, Prompt


class PromptPlugin(flow2.Task):
    def __init__(self, name, content, model=None, parser=None, **kwargs):
        self.prompt = Prompt(content=content)
        self.model = Model(model)
        self.parser = parser
        kwargs.setdefault('output', False)
        super().__init__(name, **kwargs)

    async def run(self, context=None):
        self.prompt.context.update(context or dict())
        return await self.model(self.prompt, parser=self.parser)

    def output_result(self, result):
        cli2.print(getattr(result, 'result', result))



"""
Prompt management in text files.
"""

import os

from pathlib import Path
import jinja2
import template2
import yaml

from cli2.file import File


class Template2(template2.Template2):
    def __init__(self, plugins, paths=None, **options):
        paths = paths or []
        paths += Prompt.paths()
        super().__init__(plugins, paths, **options)


class Prompt(File):
    extension = 'txt'
    PATH = '.cli2/prompts'
    USER_PATH_ENV = 'PROMPT2_USER_PATH'
    LOCAL_PATH_ENV = 'PROMPT2_LOCAL_PATH'
    paths_ep = os.getenv('PROMPT2_PATHS_EP', 'prompt2_paths')
    package_path = Path(__file__).parent / 'templates'

    def __init__(self, path=None, content=None, **context):
        self.context = context
        super().__init__(path, content)

    @property
    def metadata(self):
        env = jinja2.Environment()
        tokens = env.lex(self.content)
        for lineno, token_type, value in tokens:
            if token_type == 'comment':
                try:
                    return yaml.safe_load(value)
                except:  # noqa
                    continue
        return dict()

    async def render(self):
        return await Template2.factory().render(
            self.content,
            **self.context,
        )

    async def messages(self):
        messages = [
            dict(
                role='user',
                content=await self.render(),
            ),
        ]
        return messages



import litellm
from prompt2.plugin import Plugin
import os
import cli2
import sys


class LiteLLMPlugin(Plugin):
    def __init__(self, model_name, **model_kwargs):
        self.model_name = model_name
        self.model_kwargs = model_kwargs

    async def completion(self, messages, stream=False):
        if os.getenv('LITELLM_DEBUG'):
            litellm._turn_on_debug()
        stream = await litellm.acompletion(
            messages=messages,
            stream=True,
            model=self.model_name,
            **self.model_kwargs,
        )

        full_content = ""
        printed_lines = 0
        full_reasoning = ''
        reasoning_printed = False
        code_open = False
        async for chunk in stream:
            if hasattr(chunk, 'choices') and chunk.choices:
                delta = chunk.choices[0].delta
                if reasoning := getattr(delta, 'reasoning_content', None):
                    if stream:
                        if not reasoning_printed:
                            print(cli2.t.o.b('REASONING'), file=sys.stderr)
                            reasoning_printed = True
                        print(
                            cli2.t.G(delta.reasoning_content),
                            end='',
                            flush=True,
                            file=sys.stderr,
                        )
                    full_reasoning += reasoning

                if content := getattr(delta, 'content', ''):
                    if reasoning_printed:
                        # separate reasoning output visually
                        print('\n', file=sys.stderr)
                        reasoning_printed = False

                    full_content += content
                    if not content.endswith('\n'):
                        continue

                    new_lines = full_content.split('\n')[printed_lines:]
                    for new_line in new_lines:
                        if new_line.strip().startswith('```'):
                            code_open = not code_open

                    if not new_lines:
                        continue

                    highlight_content = full_content
                    if code_open:
                        # manuall close code block for pygments to highlight
                        if not highlight_content.endswith('\n'):
                            highlight_content += '\n'
                        highlight_content += '```'

                    highlighted = cli2.highlight(highlight_content, 'Markdown')
                    highlighted_lines = highlighted.split('\n')

                    if code_open:
                        highlighted_lines = highlighted_lines[:-1]

                    print(
                        '\n'.join(highlighted_lines[printed_lines:]),
                        flush=True,
                        file=sys.stderr,
                    )
                    printed_lines = len(highlighted_lines)

        new_lines = full_content.split('\n')[printed_lines:]
        for new_line in new_lines:
            if new_line.strip().startswith('```'):
                code_open = not code_open

        highlight_content = full_content
        if code_open:
            # manuall close code block for pygments to highlight code
            if not highlight_content.endswith('\n'):
                highlight_content += '\n'
            highlight_content += '```'

        highlighted = cli2.highlight(highlight_content, 'Markdown')
        highlighted_lines = highlighted.split('\n')

        if code_open:
            highlighted_lines = highlighted_lines[:-1]

        print(
            '\n'.join(highlighted_lines[printed_lines:]),
            flush=True,
            file=sys.stderr,
        )
        printed_lines = len(highlighted_lines)

        return full_content or full_reasoning

    def __str__(self):
        return (
            f'litellm {self.model_name} '
            + ' '.join([
                f'{k}={v}' for k, v in self.model_kwargs.items()
            ])
        ).strip()



"""
Template based prompts on CLI
"""


from cli2.file import FileCommand, FileCommands
import cli2
import importlib.metadata
import inspect
import jinja2
import prompt2


class ModelParserCommand(cli2.Command):
    def __init__(self, *args, **kwargs):
        self._model = None
        self._parser = None
        self.not_found_error = None
        super().__init__(*args, **kwargs)

    def model_parse(self):
        # auto cast variables dependending on each other ...
        if 'model' in self:
            if self['model'].value:
                configuration = prompt2.Model.configuration_get(
                    self['model'].value,
                    strict=True,
                )
                backend = prompt2.Model.backend_factory(configuration)
                self['model'].value = prompt2.Model(backend)
            else:
                self['model'].value = prompt2.Model()

    def parse(self, *argv):
        error = super().parse(*argv)
        if error:
            return error
        self.model_parse()

    def call(self, *args, **kwargs):
        if self.not_found_error:
            return self.handle_not_found(self.not_found_error)

        try:
            return super().call(*args, **kwargs)
        except prompt2.NotFoundError as exc:
            return self.handle_not_found(exc)


cli = cli2.Group('prompt2')


@cli.cmd(color='yellow', cls=ModelParserCommand)
async def ask(*args, parser=None, model=None, _cli2=None):
    """
    Ask a question from the CLI

    Example:

        prompt2 ask write a hello world in python

    :param args: Question to ask
    :param parser: Parser name if any
    :param model: Model name to use, if any
    """
    if not args:
        return _cli2.help(error='Ask a question please I am begging you!')
    await model(
        prompt2.Prompt(
            content=' '.join(args)
        ),
        parser,
    )


class PromptCommand(FileCommand, ModelParserCommand):
    file_cls = prompt2.Prompt
    file_arg = 'prompt'

    def parser_parse(self):
        if 'parser' in self:
            if self['parser'].value:
                self['parser'].value = (
                    self['model'].value.parser(self['parser'].value)
                )
            elif 'prompt' in self and self['prompt'].value:
                name = self['prompt'].value.metadata.get('parser', None)
                if name:
                    self['parser'].value = (
                        self['model'].value.parser(name)
                    )

    def parse(self, *argv):
        error = super().parse(*argv)
        if error:
            return error
        self.parser_parse()

    def handle_exception(self, exc):
        if isinstance(exc, jinja2.UndefinedError):
            prompt = self['prompt'].value
            print(
                cli2.t.bold('USED PROMPT: ')
                + cli2.t.green(f'{prompt.path}')
                + '\n'
            )
            print(cli2.t.bold('CONTENT:'))
            print(cli2.highlight(prompt.content, 'SqlJinja'))
            print()
            print(cli2.t.red.bold('MISSING CONTEXT VARIABLE'))
            print(exc.message)
        else:
            super().handle_exception(exc)


class PromptCommands(FileCommands):
    def __init__(self):
        super().__init__(prompt2.Prompt, lexer='SqlJinja')

    @cli2.cmd(cls=PromptCommand)
    def edit(self, name, local: bool = False):
        """
        Edit a prompt.

        :param name: prompt name.
        :param local: Enable this to store in $CWD/.prompt2 instead of
                      $HOME/.prompt2
        """
        return super().edit(name, local)

    @cli2.cmd(color='green', cls=PromptCommand)
    def show(self, prompt):
        """
        Show a prompt

        :param prompt: Prompt name
        """
        print(cli2.t.y.bold('PATH'))
        print(cli2.t.orange(prompt.path))
        print()

        if prompt.metadata:
            print(cli2.t.y.bold('METADATA'))
            cli2.print(prompt.metadata)
            print()

        print(cli2.t.y.bold('CONTENT'))
        print(prompt.content)

    @cli2.cmd(color='green', cls=PromptCommand)
    async def render(self, prompt, **context):
        """
        Render a prompt with a given template context.

        :param prompt: prompt name
        :param context: Context variables.
        """
        return super().render(prompt, **context)

    @cli2.cmd(color='green', cls=PromptCommand)
    async def messages(self, prompt, parser=None, model=None, **context):
        """
        Render prompt messages with a given template context.

        :param name: Prompt name
        :param parser: Parser name if any
        :param model: Model name to use, if any
        :param context: Context variables.
        """
        messages = await prompt.messages()
        if parser:
            messages = parser.messages(messages)

        print(cli2.t.y.bold('PATH'))
        print(cli2.t.orange(prompt.path))
        print()
        print(cli2.t.y.bold('OUTPUT'))
        return messages

    @cli2.cmd(cls=PromptCommand, color='yellow')
    async def send(self, prompt, parser=None, model=None, **context):
        """
        Send a prompt, rendered with a context, on a model with a parser.

        :param prompt: Prompt name
        :param parser: Parser name if any
        :param model: Model name to use, if any
        :param context: Context variables
        """
        return await model(prompt, parser)


cli.load(PromptCommands())


class ParserCommands:
    @cli2.cmd(color='gray', cls=ModelParserCommand)
    def parser(self, name):
        """
        Show a parser

        :param name: Parser name to display
        """
        parser = prompt2.Parser.get(name)

        path = inspect.getfile(parser)
        print(cli2.t.y.bold('PATH'))
        print(cli2.t.o(path))
        print()

        source = inspect.getsource(parser)
        print(cli2.highlight(source, 'Python'))

    @cli2.cmd(color='gray')
    def parsers(self):
        """
        List registered parsers
        """
        return {
            plugin.name: plugin.value
            for plugin in importlib.metadata.entry_points(
                group=prompt2.Parser.entry_point,
            )
        }


cli.load(ParserCommands())


@cli.cmd(color='gray')
def plugins():
    """
    List registered backend plugins
    """
    return {
        plugin.name: plugin.value
        for plugin in importlib.metadata.entry_points(
            group=prompt2.Plugin.entry_point,
        )
    }
